import websocket

from zymbit.settings import CHECK_HOSTNAME, WEBSOCKET_ENDPOINT, WEBSOCKET_SEND_CLIENT_INFO
from zymbit.upstream.api import ZymbitApi
from zymbit.util.client import get_client_info


def get_websocket():
    sslopt = {"check_hostname": CHECK_HOSTNAME}
    url = get_websocket_url()

    ws = websocket.create_connection(url, sslopt=sslopt)
    ws.settimeout(0)

    return ws


def get_websocket_url():
    params = {}

    if WEBSOCKET_SEND_CLIENT_INFO:
        params.update(get_client_info())

    api = ZymbitApi()
    response = api.get(WEBSOCKET_ENDPOINT, params=params)

    return response.json()



from __future__ import absolute_import

import datetime
import inspect
import logging
import time

from .time import LONG_TIME_AGO, now, get_sleep_time

NO_SLEEP = '-- NO SLEEP --'


class StateMachine(object):
    transitions = {}

    def __init__(self, raise_exceptions=True):
        self._run = True
        self._state = self.start

        self.raise_exceptions = raise_exceptions
        self.loop_sleep_time = 1.0

        self.last_exception = None

        self._setup_transitions()
        self.logger.debug('transitions={}'.format(self.transitions))

        self.check_start = False
        self.last_start = LONG_TIME_AGO
        self.next_start = LONG_TIME_AGO
        self.start_fail_count = 0
        self.start_success_delta = datetime.timedelta(seconds=10)

    def _setup_transitions(self):
        # convert the transition functions into bound methods
        _transitions = {}
        for k, v in list(self.transitions.items()):
            bound_method = getattr(self, k.__name__)
            t_transitions = dict([(kk, getattr(self, vv.__name__)) for kk, vv in list(v.items())])

            _transitions[bound_method] = t_transitions

        self.transitions = _transitions

    @property
    def logger(self):
        return logging.getLogger('{}.{}'.format(__name__, self.__class__.__name__))

    def loop(self):
        result = None

        try:
            result = self._state()
        except Exception as exc:  # global exception catcher here to use for state transitions
            self.last_exception = exc

            result = exc
            if not inspect.isclass(exc):
                result = exc.__class__

            if self.raise_exceptions:
                raise
            else:
                self.logger.exception(exc)
        else:
            self.last_exception = None
        finally:
            transitions = self.transitions.get(self._state, {})

            for _result, _state in list(transitions.items()):
                if _result == result or inspect.isclass(_result) and inspect.isclass(result) and issubclass(result, _result):
                    self._state = _state

        return result

    def quit(self):
        self._run = False

    def run(self):
        while self._run:
            start = now()

            current_state = self._state
            result = self.loop()

            # only sleep when there is no state transition
            if current_state == self._state and result != NO_SLEEP:
                sleep_time = get_sleep_time(self.loop_sleep_time, start)
                # self.logger.debug('loop_sleep_time={}, sleep_time={}'.format(self.loop_sleep_time, sleep_time))
                time.sleep(sleep_time)

    def start(self):
        _now = now()

        if self.check_start:
            self.check_start = False

            if _now > self.last_start + self.start_success_delta:
                # the current time is greater than the last start time + the
                # success delta; reset the fail count
                self.start_fail_count = 0
            else:
                # otherwise, increment the fail count and calculate an exponential
                # backoff
                self.start_fail_count += 1

                backoff = datetime.timedelta(seconds=2 ** self.start_fail_count)
                self.next_start = _now + backoff

                self.logger.info('next start at {}'.format(self.next_start))

        if _now < self.next_start:
            # the current time is before the next start, hold off
            return False

        self.check_start = True
        self.last_start = _now

        return True



import os

from email.parser import Parser

PKG_INFO_FILENAME = 'PKG-INFO'
SOURCES_FILENAME = 'SOURCES.txt'


def get_egg_info(path):
    for item in os.listdir(path):
        if item.endswith('.egg-info'):
            yield os.path.join(path, item)


def get_sources(path):
    with open(os.path.join(path, SOURCES_FILENAME), 'rb') as fh:
        return fh.read().strip().splitlines()


def find_package():
    sep = os.path.sep
    source_path = sep.join(__file__.split(sep)[-3:]).replace('.pyc', '.py')

    dirname = os.path.dirname(__file__)
    if dirname == '':
        dirname = os.getcwd()

    dirname = os.path.abspath(dirname)

    while dirname != '/':
        for item in get_egg_info(dirname):
            if source_path in get_sources(item):
                return item

        dirname = os.path.dirname(dirname)


def get_version():
    package = find_package()
    if package is None:
        return 'dev'

    pkg_info_path = os.path.join(package, PKG_INFO_FILENAME)

    with open(pkg_info_path, 'rb') as fh:
        parsed = Parser().parse(fh)

    return parsed.get('Version')



import collections


class BufferIterator(collections.Iterable):
    def __init__(self, buf=None, split_at='\n'):
        self.buf = ''
        if buf:
            self.write(buf)

        self.split_at = split_at

    def __iter__(self):
        return self

    def __next__(self):
        try:
            idx = self.buf.index(self.split_at)
        except ValueError:
            raise StopIteration()

        buf = self.buf[:idx+1]
        self.buf = self.buf[idx+1:]

        return buf

    def write(self, buf):
        try:
            buf = buf.decode('utf8')
        except AttributeError:
            pass

        self.buf += buf



import hashlib
import json
import os
import re

from zymbit.settings import AUTH_ROOT, AUTH_TOKEN, CLIENT_ID_VERSION, ZYMBIT_HOST_INFO_PATH

MAC_RE = re.compile(r'.*HWaddr (?P<hwaddr>[^ ]+)')
SDCARD_ATTRS_RE = re.compile(r'ATTRS{(?P<key>[^}]+)}=="(?P<value>[^"]+)"')


def get_auth_path():
    client_id = get_client_id()
    return os.path.join(AUTH_ROOT, client_id)


def get_auth_token():
    auth_token = AUTH_TOKEN
    if auth_token is not None:
        return auth_token

    auth_path = get_auth_path()

    if os.path.exists(auth_path):
        with open(auth_path, 'rb') as fh:
            auth_token = fh.read().strip()

    return auth_token


def get_cpu_info():
    """
    Returns CPU identification information
    :return:
    """
    info = {
        'cpu_hardware': None,
        'cpu_revision': None,
        'cpu_serial': None,
    }

    with open(os.path.join(ZYMBIT_HOST_INFO_PATH, 'cpu')) as fh:
        content = fh.read()

    for line in content.splitlines():
        line = line.strip()
        if line == '':
            continue

        line_split = line.split(':', 1)
        key = 'cpu_{}'.format(line_split[0].strip().replace(' ', '_').lower())

        if key not in list(info.keys()):
            continue

        info[key] = line_split[1].strip()

    return info


def get_eth0_info():
    """
    Returns eth0 identification information
    :return:
    """
    info = {
        'eth0_hwaddr': None
    }

    with open(os.path.join(ZYMBIT_HOST_INFO_PATH, 'eth0')) as fh:
        content = fh.read()

    for line in content.splitlines():
        matches = MAC_RE.match(line)
        if not matches:
            continue

        info['eth0_hwaddr'] = matches.group('hwaddr')

    return info


def get_sdcard_info():
    """
    Returns sdcard identification information
    :return dict: sdcard information
    """
    info = {
        'sdcard_manfid': None,
        'sdcard_serial': None,
    }

    with open(os.path.join(ZYMBIT_HOST_INFO_PATH, 'sdcard')) as fh:
        content = fh.read()

    for line in content.splitlines():
        matches = SDCARD_ATTRS_RE.match(line.strip())
        if not matches:
            continue

        key = 'sdcard_{}'.format(matches.group('key'))
        if key not in list(info.keys()):
            continue

        info[key] = matches.group('value')

    return info


def get_client_id():
    if CLIENT_ID_VERSION is None:
        return get_client_id_latest()

    return globals()['get_client_id_v{}'.format(CLIENT_ID_VERSION)]()


def get_client_id_v0():
    info = get_eth0_info()

    return info['eth0_hwaddr']


def get_client_id_v1():
    info = get_client_info()

    # the client_id is the hash of a JSON representation of an array of (key, value) 2-tuples
    data = json.dumps(sorted(list(info.items()), key=lambda a: a[0])).encode('utf8')
    sha = hashlib.sha1(data)

    return sha.hexdigest()


# alias the default get_client_id to v1
get_client_id_latest = get_client_id_v1


def get_client_info():
    info = {}

    info.update(get_cpu_info())
    info.update(get_eth0_info())
    info.update(get_sdcard_info())

    return info



import os
import sys

from collections import OrderedDict
from sh import uname


def find_cpuinfo(markers):
    # no support for non-linux systems
    if os.path.exists('/proc/cpuinfo'):
        content = open('/proc/cpuinfo', 'rb').read()
        for marker, return_value in list(markers.items()):
            if marker in content:
                return return_value


def get_device_meta():
    return {
        'distro': get_distro(),
        'model': get_model(),
        'system': get_system(),
    }


def get_distro():
    """
    Returns the device distribution

    :return string: device distribution
    """
    distro = None

    if os.path.exists('/etc/linino'):
        return 'linino'

    result = uname('-a')
    if 'Darwin Kernel Version' in result:
        return 'osx'

    issue_path = '/etc/issue'
    if os.path.exists(issue_path):
        with open(issue_path, 'rb') as fh:
            content = fh.read().lower()
            for _distro in ('raspbian', 'arch'):
                if _distro in content:
                    distro = _distro
                    break

    return distro


def get_model():
    systems = OrderedDict((
        ('Arduino Yun', 'yun'),
        ('BCM2708', '1'),
        ('BCM2709', '2'),
    ))
    cpuinfo = find_cpuinfo(systems)
    if cpuinfo:
        return cpuinfo

    if sys.platform == 'darwin':
        return sys.platform


def get_system():
    systems = OrderedDict((
        ('Arduino Yun', 'arduino'),
        ('BCM270', 'raspberrypi'),  # note this will match BCM2708 (rpi) and BCM2709 (rpi2)
    ))
    cpuinfo = find_cpuinfo(systems)
    if cpuinfo:
        return cpuinfo

    if sys.platform == 'darwin':
        return sys.platform

    if 'linux' in sys.platform:
        return 'linux'



import datetime
import dateutil.parser
import functools
import logging
import pytz

# it's impossible that "now" is less than this datetime
# we know we are out of sync with real time if we ever
# get a time value less than this
MIN_DT = datetime.datetime(2014, 7, 25, 17, 00, 00)  # Zymbit est date, UTC

utc = pytz.utc

EPOCH = datetime.datetime.utcfromtimestamp(0).replace(tzinfo=utc)

LONG_TIME_AGO = utc.localize(datetime.datetime(1, 1, 1))  # a really long time ago

# keys follow the same convention as InfluxDB
SECOND_PRECISIONS = {
    's': 1,
    'ms': 1000,
    'u': 1e6,
    'n': 1e9,
}


def now():
    return utc.localize(datetime.datetime.utcnow())


def timestamp(dt=None):
    if dt is None:
        dt = now()

    return dt.isoformat('T')


def get_sleep_time(seconds, start):
    """
    Wait at most the given number of seconds from the initial time given
    :param seconds: float - number of seconds to wait
    :param start: datetime - the start time
    :return: float - time to wait
    """
    _now = now()
    delta = _now - start
    diff = delta.seconds + (1.0 * delta.microseconds / 1e6)
    wait = max(0, seconds - diff)

    # print 'start={}, _now={}, delta={}, diff={}, wait={}'.format(start, _now, delta, diff, wait)

    return wait


def interval(interval_delay, default_return=None):
    """
    Call a function every given interval
    :param interval_delay: float - number of seconds
    :param default_return: when the interval has not passed, what to return (default: None)
    """
    interval_delta = datetime.timedelta(seconds=interval_delay)

    def wrapper(fn):
        @functools.wraps(fn)
        def interval_handler(*args, **kwargs):
            t0 = now()
            last_call = getattr(fn, 'last_call', LONG_TIME_AGO)

            if (t0 - last_call) > interval_delta:
                fn.last_call = t0

                return fn(*args, **kwargs)
            else:
                return default_return

        return interval_handler

    return wrapper


class MillisDatetime(object):
    def __init__(self, millis):
        self.last_millis = None
        self.initial = None

        self.set_initial(millis)

    @property
    def logger(self):
        return logging.getLogger(__name__)

    def get_time(self, millis):
        if millis < self.last_millis:
            self.logger.info(
                'time rolled over, last_millis={}, millis={}'.format(
                    self.last_millis, millis))

            self.set_initial(millis)

        delta = datetime.timedelta(milliseconds=millis)
        return self.initial + delta

    def set_initial(self, millis):
        delta = datetime.timedelta(milliseconds=millis)
        self.initial = now() - delta

        self.last_millis = millis


def get_seconds(iso_timestamp, precision='s'):
    """
    Returns the number of seconds since EPOCH for the given ISO 8601 timestamp
    """
    dt = dateutil.parser.parse(iso_timestamp)

    return get_seconds_dt(dt, precision=precision)


def get_seconds_dt(dt=None, precision='s'):
    """
    Returns the number of seconds since EPOCH for the given datetime object
    """
    dt = dt or now()

    return (dt - EPOCH).total_seconds() * SECOND_PRECISIONS[precision]



#!/usr/bin/env python
import os

from distutils.core import setup

SCRIPT_DIR = os.path.dirname(__file__)
if not SCRIPT_DIR:
        SCRIPT_DIR = os.getcwd()

# put together list of requirements to install
install_requires = []
with open(os.path.join(SCRIPT_DIR, 'requirements.txt')) as fh:
    for line in fh.readlines():
        if line.startswith('-'):
            continue

        install_requires.append(line.strip())

data_files = []

setup(name='zymbit-connect',
      version='0.4.3',
      description='Zymbit Connect',
      author='Roberto Aguilar',
      author_email='roberto@zymbit.com',
      package_dir={'': 'src'},
      packages=[
          'zymbit',
          'zymbit.connect',
          'zymbit.upstream',
          'zymbit.util',
      ],
      scripts=[
          'src/scripts/connect',
          'src/scripts/write_auth_token',
      ],
      data_files=data_files,
      long_description=open('README.md').read(),
      url='http://zymbit.com/',
      license='LICENSE',
      install_requires=install_requires,
)



import logging
import socket

from select import select

from zymbit import settings
from zymbit.util.envelope import get_envelope

BUFSIZE = 4096

# backwards compat with python2
try:
    BlockingIOError
except NameError:
    BlockingIOError = None.__class__

try:
    ConnectionResetError
except NameError:
    ConnectionResetError = None.__class__


class BaseServer(object):
    def __init__(self, host, port, message_handler=None):
        self.addr = (host, port)

        self._tcp_sock = None
        self._udp_sock = None

        self.connections = {}

        self.message_handler = message_handler

        self._run = True

    @property
    def logger(self):
        logger_name = '{}.{}'.format(__name__, self.__class__.__name__)
        return logging.getLogger(logger_name)

    @property
    def tcp_sock(self):
        if self._tcp_sock:
            return self._tcp_sock

        try:
            self._tcp_sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

            self._tcp_sock.setblocking(0)
            self._tcp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._tcp_sock.bind(self.addr)
            self._tcp_sock.listen(128)  # max 128 clients
        except socket.error:
            self.logger.warning('Unable to bind TCP socket at addr={}'.format(self.addr))
        else:
            self.logger.info("Listening on TCP addr={}".format(self.addr))

        return self._tcp_sock

    @property
    def udp_sock(self):
        if self._udp_sock:
            return self._udp_sock

        try:
            self._udp_sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)

            self._udp_sock.setblocking(0)
            self._udp_sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            self._udp_sock.bind(self.addr)
        except socket.error:
            self.logger.warning('Unable to bind UDP socket at addr={}'.format(self.addr))
        else:
            self.logger.info("Listening on UDP addr={}".format(self.addr))

        return self._udp_sock

    def broadcast(self, message):
        for connection in self.connections:
            self.send(connection, message)

    def close_tcp(self):
        self._tcp_sock = None

    def close_udp(self):
        self._udp_sock = None

    def connect(self, info):
        message = get_envelope('connection', dict(routing_key='connection.connected'))

        conn, addr = info
        self.logger.info('%s, %s %s' % (conn, addr, message))

        self.connections[conn] = addr

        self.handle_message(conn, message)

    def disconnect(self, connection):
        message = get_envelope('connection', dict(routing_key='connection.disconnected'))

        addr = self.connections.pop(connection)
        self.logger.info('%s, %s %s' % (connection, addr, message))

        self.handle_message(connection, message)

    def fileno(self):
        return self.tcp_sock.fileno()

    def handle_message(self, client, buf):
        if self.message_handler:
            self.message_handler(client, buf)
        else:
            self.logger.info('client={}, buf={}'.format(client, buf))

    def loop(self, select_timeout=1.0):
        handled = None

        # check UDP
        try:
            buf, client = self.udp_sock.recvfrom(1024)
        except socket.error as exc:
            if isinstance(exc, (BlockingIOError,)):
                error_number = exc.errno
            else:
                error_number = exc[0]

            # (11, 'Resource temporarily unavailable')
            # [Errno 35] Resource temporarily unavailable
            if error_number not in (11, 35):
                self.logger.exception(exc)
                self.logger.warning('got socket error_number={}'.format(error_number))

                self.close_udp()
        else:
            if buf:
                self.handle_message(client, buf)
                handled = True

        try:
            self.connect(self.tcp_sock.accept())
        except socket.error as exc:
            if isinstance(exc, (BlockingIOError,)):
                error_number = exc.errno
            else:
                error_number = exc[0]

            # (11, 'Resource temporarily unavailable')
            # [Errno 35] Resource temporarily unavailable
            if error_number not in (11, 35):
                self.logger.exception(exc)
                self.logger.warning('got socket error_number={}'.format(error_number))

                self.close_tcp()

        ready, _, _ = select(self.connections, [], [], select_timeout)

        for client in ready:
            try:
                buf = client.recv(BUFSIZE)
            except socket.error as exc:
                if isinstance(exc, (ConnectionResetError,)):
                    error_number = exc.errno
                else:
                    error_number = exc[0]

                # [Errno 54] Connection reset by peer
                # [Errno 104] Connection reset by peer -- raspbian
                if error_number not in (54, 104):
                    self.logger.exception(exc)
                    self.logger.warning('got socket error_number={}'.format(error_number))

                self.disconnect(client)
                continue
            else:
                if not len(buf):
                    self.disconnect(client)
                    continue

            self.handle_message(client, buf)
            handled = True

        return handled

    def quit(self):
        self.tcp_sock.close()
        self.udp_sock.close()

        # prevent getting exception where dictionary changes while looping
        connections = list(self.connections.keys())
        for connection in connections:
            self.disconnect(connection)

    def run(self):
        while self._run:
            self.loop()

    def send(self, connection, buf):
        try:
            if not isinstance(buf, (bytes,)):
                buf = buf.encode('utf8')

            connection.send(buf)
        except Exception as exc:
            self.logger.exception(exc)
            self.logger.error('error sending connection={}, buf={}'.format(connection, buf))


class ConsoleMessengerServer(BaseServer):
    def __init__(self, message_handler):
        super(ConsoleMessengerServer, self).__init__(
                settings.CONSOLE_MESSENGER_HOST,
                settings.CONSOLE_MESSENGER_PORT,
                message_handler=message_handler
        )
